package db.service;

import db.Globals;
import db.entity.LogFileMeta;
import db.entity.LogItem;
import db.mapper.LogFileMetaMapper;
import db.util.ConvertUtils;
import db.util.FileUtils;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.util.*;

@Service
@Data
@Slf4j
public class LogFileService {

    @Autowired
    private LogFileMetaMapper logFileMetaMapper;

    @Value("${system.data-file.prefix}")
    private String prefix;
    @Value("${system.data-file.suffix}")
    private String suffix;
    @Value("${system.data-file.expire-time}")
    private Integer expireHour;
    @Value("${system.data-file.retry-time}")
    private Integer retryMillisecond;

    private LogFileMeta logFileMeta; // 当前应写入的文件头信息(用于单logItem写入，废弃)

    public byte[] queryLog(long loc) throws IOException {
        log.debug("读取{}", loc);
        int fileIndex = (int) (loc % Globals.MAX_LOG_ITEM);
        long fileId = loc / Globals.MAX_LOG_ITEM;

        // 读取fileId文件的头
        LogFileMeta readLogFile = logFileMetaMapper.getLogFileMetaById(fileId);

        byte[] headerContent = new byte[Globals.MAX_LOG_ITEM * 4 + 28];
        FileUtils.readFile(readLogFile.getFileName(), headerContent, 0);
        boolean isReadSuccess = readLogFile.setPayload(headerContent);
        if (!isReadSuccess) {
            log.error("logFile文件完整性校验失败，文件名:{}", readLogFile.getFileName());
            return null;
        }
        int fileOffset = readLogFile.getIndices()[fileIndex];

        // 读取logItem的header
        byte[] itemHeaderContent = new byte[16];
        FileUtils.readFile(readLogFile.getFileName(), itemHeaderContent, fileOffset);
        LogItem logItem = new LogItem();

        isReadSuccess = logItem.setHeader(itemHeaderContent);
        ;
        if (!isReadSuccess) {
            log.error("logItem文件完整性校验失败，文件名:{},文件偏移量：{}", readLogFile.getFileName(), fileOffset);
            return null;
        }

        // 读取logItem的body
        byte[] itemContent = new byte[logItem.getLength() - 16];
        FileUtils.readFile(readLogFile.getFileName(), itemContent, fileOffset + 16);
        return itemContent;
    }

    public List<byte[]> queryLogBetweenTime(LogFileMeta logFileMeta, long startTime, long endTime) throws IOException {
        List<byte[]> res = new ArrayList<>();

        // 读取fileId文件的头
        LogFileMeta readLogFile = logFileMetaMapper.getLogFileMetaById(logFileMeta.getId());

        byte[] headerContent = new byte[Globals.MAX_LOG_ITEM * 4 + 28];
        FileUtils.readFile(readLogFile.getFileName(), headerContent, 0);
        boolean isReadSuccess = readLogFile.setPayload(headerContent);
        if (!isReadSuccess) {
            log.error("logFile文件完整性校验失败，文件名:{}", readLogFile.getFileName());
            return null;
        }

        // 遍历索引
        for (int i = 0; i < logFileMeta.getCount(); i++) {
            int fileOffset = readLogFile.getIndices()[i];
            // 读取logItem的header
            byte[] itemHeaderContent = new byte[16];
            FileUtils.readFile(readLogFile.getFileName(), itemHeaderContent, fileOffset);
            LogItem logItem = new LogItem();
            isReadSuccess = logItem.setHeader(itemHeaderContent);
            ;
            if (!isReadSuccess) {
                log.error("logItem文件完整性校验失败，文件名:{},文件偏移量：{}", readLogFile.getFileName(), fileOffset);
                return null;
            }
            if (logItem.getStamp() >= startTime && logItem.getStamp() <= endTime) {
                // 读取logItem的body
                byte[] itemContent = new byte[logItem.getLength() - 16];
                FileUtils.readFile(readLogFile.getFileName(), itemContent, fileOffset + 16);
                res.add(itemContent);
            }
        }
        return res;
    }

    /**
     * 在关系数据库中标记指定 logFile 存入 TSDB 成功
     *
     * @param logFileId
     */
    public void markTsdbSuccess(long logFileId) {
        // 查找对应文件
        LogFileMeta logFileMeta = logFileMetaMapper.getLogFileMetaById(logFileId);
        logFileMeta.setSaveTsdb(1);
        logFileMeta.setSaveTsdbStamp(System.currentTimeMillis());
        logFileMetaMapper.update(logFileMeta);
    }

    public Long getWalStamp(long logFileId) {
        // 查找对应文件
        LogFileMeta logFileMeta = logFileMetaMapper.getLogFileMetaById(logFileId);
        return logFileMeta.getWalStamp();
    }

    /**
     * 在关系数据库中标记指定 logFile 尝试存入 TSDB 时间
     *
     * @param logFileId
     */
    public void markTsdbTryStamp(long logFileId) {
        // 查找对应文件
        LogFileMeta logFileMeta = logFileMetaMapper.getLogFileMetaById(logFileId);
        logFileMeta.setSaveTsdbStamp(System.currentTimeMillis());
        logFileMetaMapper.update(logFileMeta);
    }

    /**
     * 插入多条(不多于MaxItem条)数据，返回值为Long二进制转的Double
     *
     * @param logItems
     * @param totalRawPacketSize
     * @return
     * @throws IOException
     */
    public List<Double> appendLogs(List<LogItem> logItems, int totalRawPacketSize, long walStamp) throws IOException {
        if(logItems.size() < Globals.MAX_LOG_ITEM) {
            return appendLogsLessThanOneFile(logItems, totalRawPacketSize, walStamp);
        } else if(logItems.size() == Globals.MAX_LOG_ITEM) {
            return appendLogsEqualThanOneFile(logItems, totalRawPacketSize, walStamp);
        } else {
            // 不应该到这里
            log.warn("[appendLogs] logItems.size() > Globals.MAX_LOG_ITEM");
            return null;
        }
    }

    /**
     * 插入少于MaxItem条数据
     * @param logItems
     * @param totalRawPacketSize
     * @return
     * @throws IOException
     */
    private List<Double> appendLogsLessThanOneFile(List<LogItem> logItems, int totalRawPacketSize, long walStamp)
            throws IOException {
        // 查找未满的文件名
        LogFileMeta logFileMeta = findUnfilledLogFile(logItems.size());

        if (logFileMeta == null) {
            // 无剩余的空间，写入新文件
            return appendLogsEqualThanOneFile(logItems, totalRawPacketSize, walStamp);
        }

        // 记录尝试更新时间
        logFileMeta.setSaveDiskStamp(System.currentTimeMillis());
        logFileMeta.setWalStamp(walStamp);
        logFileMetaMapper.update(logFileMeta);

        List<Double> res = new ArrayList<>(logItems.size());

        // 读取文件头
        byte[] headerContent = new byte[Globals.MAX_LOG_ITEM * 4 + 28];
        FileUtils.readFile(logFileMeta.getFileName(), headerContent, 0);

        boolean isReadSuccess = logFileMeta.setPayload(headerContent);
        if (!isReadSuccess) {
            log.error("logFile文件完整性校验失败，文件名:{}", logFileMeta.getFileName());
            return null;
        }

        // 寻找写入位置
        // 读取上一条记录的长度
        int fileOffset = -1;
        Integer index = logFileMeta.getCount();
        if (index == 0) {
            // 第一条
            fileOffset = headerContent.length;
            log.debug("log文件第一条，fileOffset={}", fileOffset);
        } else if (index == Globals.MAX_LOG_ITEM) {
            // 已满
            log.error("log 文件最后一条，不应该到这里");
            return null;
        } else {
            byte[] lastContent = new byte[16];
            FileUtils.readFile(logFileMeta.getFileName(), lastContent,
                    logFileMeta.getIndices()[index - 1]);
            LogItem lastItem = new LogItem();

            isReadSuccess = lastItem.setHeader(lastContent);
            if (!isReadSuccess) {
                log.error("logItem 文件完整性校验失败，文件名:{},文件偏移量：{}", logFileMeta.getFileName(),
                        logFileMeta.getIndices()[index - 1]);
                return null;
            }
            fileOffset = lastItem.getLength() + logFileMeta.getIndices()[index - 1];
            log.debug("log 文件非首尾条，fileOffset={}", fileOffset);
        }

        int bodyTotalLength = totalRawPacketSize + logItems.size() * 16; // header(16)*items

        byte[] writeBuffer = new byte[bodyTotalLength];

        int bufferOffset = 0;
        // 拼接要写入的内容
        for (int i = 0; i < logItems.size(); i++) {
            LogItem logItem = logItems.get(i);
            // bufferOffset是body部分的偏移
            res.add(Double.longBitsToDouble(logFileMeta.getId() * Globals.MAX_LOG_ITEM + i));
            byte[] payload = logItem.getPayload();
            ConvertUtils.arrayCopy(payload, 0, writeBuffer, bufferOffset, payload.length);
            logFileMeta.getIndices()[i] = logFileMeta.getHeaderTotalLength() + bufferOffset;
            bufferOffset += payload.length;
            if (i == 0) {
                logFileMeta.setMinTime(logItem.getStamp());
                logFileMeta.setMaxTime(logItem.getStamp());
            } else {
                logFileMeta.setMinTime(Math.min(logFileMeta.getMinTime(), logItem.getStamp()));
                logFileMeta.setMaxTime(Math.max(logFileMeta.getMaxTime(), logItem.getStamp()));
            }
        }

        // 更新文件元信息
        logFileMeta.setCount(logItems.size());
        logFileMeta.setBodyLength(bodyTotalLength);

        // 修改文件头
        FileUtils.overwriteFile(logFileMeta.getFileName(), logFileMeta.getPayload(), 0);

        // 写入数据
        FileUtils.appendFile(logFileMeta.getFileName(), writeBuffer);

        logFileMeta.setSaveDisk(1); // 1-true; 0-false
        logFileMeta.setSaveDiskStamp(System.currentTimeMillis());
        logFileMetaMapper.update(logFileMeta);
        return res;
    }

    /**
     * 插入等于MaxItem条数据
     * @param logItems
     * @param totalRawPacketSize
     * @return
     * @throws IOException
     */
    private List<Double> appendLogsEqualThanOneFile(List<LogItem> logItems, int totalRawPacketSize, long walStamp)
            throws IOException {

        // 申请新文件名
        LogFileMeta logFileMeta = applyNewLogFile();

        // 记录尝试更新时间
        logFileMeta.setSaveDiskStamp(System.currentTimeMillis());
        logFileMeta.setWalStamp(walStamp);
        logFileMetaMapper.update(logFileMeta);

        List<Double> res = new ArrayList<>(logItems.size());
        int offset = 0;

        int bodyTotalLength = totalRawPacketSize + logItems.size() * 16; // header(16)*items

        byte[] writeBuffer = new byte[bodyTotalLength];

        // 拼接要写入的内容
        for (int i = 0; i < logItems.size(); i++) {
            LogItem logItem = logItems.get(i);
            // offset是body部分的偏移
            res.add(Double.longBitsToDouble(logFileMeta.getId() * Globals.MAX_LOG_ITEM + i));
            byte[] payload = logItem.getPayload();
            ConvertUtils.arrayCopy(payload, 0, writeBuffer, offset, payload.length);
            logFileMeta.getIndices()[i] = logFileMeta.getHeaderTotalLength() + offset;
            offset += payload.length;
            if (i == 0) {
                logFileMeta.setMinTime(logItem.getStamp());
                logFileMeta.setMaxTime(logItem.getStamp());
            } else {
                logFileMeta.setMinTime(Math.min(logFileMeta.getMinTime(), logItem.getStamp()));
                logFileMeta.setMaxTime(Math.max(logFileMeta.getMaxTime(), logItem.getStamp()));
            }
        }

        // 更新文件元信息
        logFileMeta.setCount(logItems.size());
        logFileMeta.setBodyLength(bodyTotalLength);


        // 修改文件头
        FileUtils.overwriteFile(logFileMeta.getFileName(), logFileMeta.getPayload(), 0);

        // 写入数据
        FileUtils.appendFile(logFileMeta.getFileName(), writeBuffer);

        logFileMeta.setSaveDisk(1); // 1-true; 0-false
        logFileMeta.setSaveDiskStamp(System.currentTimeMillis());
        logFileMetaMapper.update(logFileMeta);
        return res;
    }


    /**
     * 获取一个新log文件的文件名
     *
     * @return
     */
    private LogFileMeta applyNewLogFile() throws IOException {
        // 查询db，确认存放文件名
        LogFileMeta lastLogFileMeta = logFileMetaMapper.getLastLogFileMeta();
        // 创建新文件
        long fileId;
        if (lastLogFileMeta == null) {
            fileId = 0;
        } else {
            fileId = lastLogFileMeta.getId() + 1;
        }
        String fileName = String.format("%s%d%s", prefix, fileId, suffix);
        log.info("创建新log文件，fileId:{}，fileName:{}", fileId, fileName);
        LogFileMeta logFileMeta = new LogFileMeta(fileId, fileName);
        logFileMetaMapper.insert(logFileMeta);
        createNewFile(logFileMeta);
        return logFileMeta;
    }

    /**
     * 获取一个未填充满的log文件的文件名
     *
     * @param requiredSize 所需空间
     * @return
     */
    private LogFileMeta findUnfilledLogFile(int requiredSize) throws IOException {
        LogFileMeta fillableLogFileMeta = logFileMetaMapper.getValidUnfilledLogFileMeta(Globals.MAX_LOG_ITEM - requiredSize);

        return fillableLogFileMeta;
    }

    @Deprecated
    public long appendLog(LogItem logItem) throws IOException {
        // 确定将要写入的文件名
        confirmLogFile();

        // 插入磁盘
        int fileIndex = appendLogToDisk(logItem);
        logFileMetaMapper.update(logFileMeta);
        return logFileMeta.getId() * Globals.MAX_LOG_ITEM + fileIndex;
    }

    /**
     * 查询DB，确认待存放的文件名，如果最近写过的文件未写满则为最近文件，否则为新文件
     *
     * @throws IOException
     */
    @Deprecated
    private void confirmLogFile() throws IOException {
        // 查询db，确认存放文件名
        LogFileMeta lastLogFileMeta = logFileMetaMapper.getLastLogFileMeta();
        if (lastLogFileMeta == null || lastLogFileMeta.getCount() >= Globals.MAX_LOG_ITEM) {
            // 文件已满，创建新文件
            long fileId;
            if (lastLogFileMeta == null) {
                fileId = 0;
            } else {
                fileId = lastLogFileMeta.getId() + 1;
            }
            String fileName = String.format("%s%d%s", prefix, fileId, suffix);
            log.info("旧log文件fileId:{}已满，创建新log文件fileId:{} fileName:{}", fileId - 1,
                    fileId, fileName);
            logFileMeta = new LogFileMeta(fileId, fileName);
            logFileMetaMapper.insert(logFileMeta);
            createNewFile(logFileMeta);
        } else {
            // 文件没满
            logFileMeta = lastLogFileMeta;
        }
    }

    @Deprecated
    public Integer appendLogToDisk(LogItem logItem) throws IOException {

        // 读取文件头
        byte[] headerContent = new byte[Globals.MAX_LOG_ITEM * 4 + 28];
        FileUtils.readFile(logFileMeta.getFileName(), headerContent, 0);

        boolean isReadSuccess = logFileMeta.setPayload(headerContent);
        if (!isReadSuccess) {
            log.error("logFile文件完整性校验失败，文件名:{}", logFileMeta.getFileName());
            return null;
        }

        // 寻找写入位置
        // 读取上一条记录的长度
        int offset = -1;
        Integer index = logFileMeta.getCount();
        if (index == 0) {
            // 第一条
            offset = headerContent.length;
            log.debug("log文件第一条，offset={}", offset);
        } else if (index == Globals.MAX_LOG_ITEM) {
            // 已满
            log.error("log文件最后一条，不应该到这里");
        } else {
            byte[] lastContent = new byte[16];
            FileUtils.readFile(logFileMeta.getFileName(), lastContent,
                    logFileMeta.getIndices()[index - 1]);
            LogItem lastItem = new LogItem();

            isReadSuccess = lastItem.setHeader(lastContent);
            if (!isReadSuccess) {
                log.error("logItem文件完整性校验失败，文件名:{},文件偏移量：{}", logFileMeta.getFileName(),
                        logFileMeta.getIndices()[index - 1]);
                return null;
            }
            offset = lastItem.getLength() + logFileMeta.getIndices()[index - 1];
            log.debug("log文件非首尾条，offset={}", offset);
        }

        // 写入数据
        byte[] logItemPayload = logItem.getPayload();
        FileUtils.appendFile(logFileMeta.getFileName(), logItemPayload);

        // 修改文件头
        logFileMeta.getIndices()[index] = offset;
        logFileMeta.setBodyLength(logFileMeta.getBodyLength() + logItemPayload.length);
        if (index == 0) {
            logFileMeta.setMinTime(logItem.getStamp());
            logFileMeta.setMaxTime(logItem.getStamp());
        } else {
            logFileMeta.setMinTime(Math.min(logFileMeta.getMinTime(), logItem.getStamp()));
            logFileMeta.setMaxTime(Math.max(logFileMeta.getMaxTime(), logItem.getStamp()));
        }
        logFileMeta.setCount(index + 1);
        log.debug("修改文件头索引，将{}位置的索引修改为{}", index, offset);
        FileUtils.overwriteFile(logFileMeta.getFileName(), logFileMeta.getPayload(), 0);
        return index;
    }

    public void createNewFile(LogFileMeta logFileMeta) throws IOException {
        // 创建索引部分
        FileUtils.writeFile(logFileMeta.getFileName(), logFileMeta.getPayload());
    }

    /**
     * 获取要重试保存的logFile
     */
    public List<LogFileMeta> getRetryTask() {
        Calendar c = Calendar.getInstance();
        c.setTime(new Date());
        c.add(Calendar.MILLISECOND, -retryMillisecond);
        long expireStamp = c.getTime().getTime();
        List<LogFileMeta> logFileMetas = logFileMetaMapper.getValidTimeOutLogFileMeta(expireStamp);
        return logFileMetas;
    }

    /**
     * 获取要重试保存磁盘的logFile
     */
    public List<LogFileMeta> getRetrySaveDisk() {
        Calendar c = Calendar.getInstance();
        c.setTime(new Date());
        c.add(Calendar.MILLISECOND, -retryMillisecond);
        long expireStamp = c.getTime().getTime();
        List<LogFileMeta> logFileMetas = logFileMetaMapper.getValidDiskTimeOutLogFileMeta(expireStamp);
        return logFileMetas;
    }

    /**
     * 获取要重试保存TSDB的logFile
     */
    public List<LogFileMeta> getRetrySaveTsdb() {
        Calendar c = Calendar.getInstance();
        c.setTime(new Date());
        c.add(Calendar.MILLISECOND, -retryMillisecond);
        long expireStamp = c.getTime().getTime();
        List<LogFileMeta> logFileMetas = logFileMetaMapper.getValidTsdbTimeOutLogFileMeta(expireStamp);

        return logFileMetas;
    }

    @Scheduled(fixedDelay = 1000 * 3600 * 12)
    public void deleteExpiredFile() {
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(new Date());
        calendar.add(Calendar.HOUR, -expireHour);
        long expireTime = calendar.getTime().getTime();
        log.debug("尝试删除过期log文件，截至时间戳为{}", expireTime);
        List<LogFileMeta> expiredLogFiles = logFileMetaMapper.getValidExpiredLogFileMeta(expireTime);
        for (LogFileMeta expiredLogFile : expiredLogFiles) {
            FileUtils.deleteFile(expiredLogFile.getFileName());
        }
        log.debug("已删除过期log文件：{}", expiredLogFiles);
        logFileMetaMapper.markValidExpiredLogFileMetaAsInvalid(expireTime);
    }

    public void markInvalidById(long logId) {
        logFileMetaMapper.markInvalidById(logId);
    }
}
