package db.handler;

import db.*;
import db.entity.LogItem;
import db.entity.Probe;
import db.entity.WalItem;
import db.entity.packages.DataResponsePackage;
import db.entity.packages.ListenResponsePackage;
import db.entity.packages.ScanResponsePackage;
import db.service.LogFileService;
import db.service.ProbeService;
import db.service.StatisticService;
import db.service.WalFileService;
import db.util.SpentTimeCalculator;
import db.util.Util;
import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;
import org.springframework.util.ObjectUtils;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.*;

@Component
@Slf4j
public class PackageHandler {

    @Value("${reso-db.metric-name}")
    private String metricName;

    @Autowired
    private DBApiEntry entry;

    @Autowired
    private LogFileService logFileService;

    @Autowired
    private ProbeService probeService;

    @Autowired
    private StatisticService statisticService;

    @Autowired
    private WalFileService walFileService;

    @Value("${system.data-file.retry-time}")
    private long retryTaskDelayMilli;

    @Value("${system.wal-file.enable}")
    private boolean enableWal;

    private long bufferIndex = 0;
    private SpentTimeCalculator bufferFullTimeCalculator = SpentTimeCalculator.create().begin();

    private List<DataResponsePackage> packageBuffer = new ArrayList<>(Globals.MAX_LOG_ITEM);

    private DelayQueue<TsdbTask> tsdbTaskBuffer = new DelayQueue<>();
    private DelayQueue<DiskTask> diskTaskBuffer = new DelayQueue<>();

    @Data
    @AllArgsConstructor
    class TsdbTask implements Delayed {
        private List<DBVal> dbVals;
        private Long bufferId;
        private Long logId;
        private Long execTime; // 执行时刻

        @Override
        public long getDelay(TimeUnit unit) {
            return execTime - System.currentTimeMillis();
        }

        @Override
        public int compareTo(Delayed o) {
            TsdbTask task = (TsdbTask) o;
            long diff = this.execTime - task.execTime;
            if (diff <= 0) {
                return -1;
            } else {
                return 1;
            }
        }
    }

    @Data
    @AllArgsConstructor
    class DiskTask implements Delayed {
        private List<DataResponsePackage> packages;
        private Long bufferId;
        private Long walId;
        private Long execTime; // 执行时刻

        @Override
        public long getDelay(TimeUnit unit) {
            return execTime - System.currentTimeMillis();
        }

        @Override
        public int compareTo(Delayed o) {
            DiskTask task = (DiskTask) o;
            long diff = this.execTime - task.execTime;
            if (diff <= 0) {
                return -1;
            } else {
                return 1;
            }
        }
    }

    @Data
    class SavePackageResult {
        private boolean saveDiskSuccess = false;
        private boolean saveTsdbSuccess = false;

        public boolean isSuccess() {
            return this.saveDiskSuccess && this.saveTsdbSuccess;
        }
    }

    /**
     * 启动时调用，检查现存的 wal 文件读取到缓冲区
     */
    @PostConstruct
    public void checkAllWal() {
        if (!enableWal) {
            log.info("未启用Wal机制");
            return;
        }
        log.info("开始检测wal文件");
        List<Long> walLists;
        try {
            walLists = walFileService.getOldWalIds();
        } catch (IOException e) {
            log.error("读取现存wal文件出错", e);
            return;
        }

        if (!ObjectUtils.isEmpty(walLists)) {
            log.info("检测到wal文件:{}，开始恢复", walLists);
            for (Long walId : walLists) {
                List<WalItem> walItems = walFileService.readWal(walId);
                replayWal(walId, walItems);
            }
        }
    }

    /**
     * 重放 wal 文件
     *
     * @param walItems
     */
    public void replayWal(long walId, List<WalItem> walItems) {
        if (walItems.size() == Globals.MAX_LOG_ITEM) {
            log.info("wal文件：{}数目达到Max_Item，直接放入磁盘任务队列", walId);
            List<DataResponsePackage> target = new ArrayList<>(Globals.MAX_LOG_ITEM);
            for (WalItem walItem : walItems) {
                DataResponsePackage dataResponsePackage = new DataResponsePackage(walItem.getProbeId(),
                        walItem.getPort(), walItem.getDataStrategy(), walItem.getClockSync(),
                        walItem.getStamp(), walItem.getRawPacket());
                target.add(dataResponsePackage);
            }
            saveFulledPackageBuffer(target, -1, walId);
            // 直接保存
        } else if (walItems.size() < Globals.MAX_LOG_ITEM) {
            // 直接写入buffer
            log.info("wal文件：{}数目不足Max_Item，直接放入缓冲区", walId);
            walFileService.openWalFile(walId);
            for (WalItem walItem : walItems) {
                DataResponsePackage dataResponsePackage = new DataResponsePackage(walItem.getProbeId(),
                        walItem.getPort(), walItem.getDataStrategy(), walItem.getClockSync(),
                        walItem.getStamp(), walItem.getRawPacket());
                addToPackageBuffer(dataResponsePackage);
            }
        } else {
            // 不应到这里
            log.warn("[replayWal] walItems.size() > Globals.MAX_LOG_ITEM");
            return;
        }

        log.info("恢复{}条wal", walItems.size());
    }

    @PostConstruct
    public void startCheckDiskQueue() {
        new Thread(() -> {
            checkDiskQueue();
        }, "CheckDiskQueueThread").start();
    }

    @PostConstruct
    public void startCheckTsdbQueue() {
        new Thread(() -> {
            checkTsdbQueue();
        }, "CheckTsdbQueueThread").start();
    }

    public void handleDataResponsePackage(DataResponsePackage dataResponsePackage) {
        log.debug("开始处理探针上传的数据包:{}", dataResponsePackage);
        statisticService.incrDataCount();

        if (enableWal) {  // 生成 WAL 文件
            WalItem walItem = new WalItem(dataResponsePackage.getProbeId(), dataResponsePackage.getPort(),
                    dataResponsePackage.getDataStrategy(), dataResponsePackage.getClockSync(),
                    dataResponsePackage.getTimestamp(), dataResponsePackage.getRawPacket());

            try {
                walFileService.appendWal(walItem);
            } catch (IOException e) {
                log.error("添加walItem到wal文件失败,walItem:{}", walItem, e);
            }
        }
        addToPackageBuffer(dataResponsePackage);
    }

    /**
     * 添加数据报文到缓冲区
     *
     * @param dataResponsePackage
     */
    public void addToPackageBuffer(DataResponsePackage dataResponsePackage) {
        // 添加至缓冲区
        packageBuffer.add(dataResponsePackage);

        if (packageBuffer.size() >= Globals.MAX_LOG_ITEM) {
            // buffer满，记录满的时间
            bufferFullTimeCalculator.end();

            // 新建缓冲区
            List<DataResponsePackage> fulledBuffer = packageBuffer;
            packageBuffer = new ArrayList<>(Globals.MAX_LOG_ITEM);
            Long walId;
            if (enableWal) {
                walId = walFileService.getAppendableWalId();
                walFileService.closeWalFile();
            } else {
                walId = -1L;
            }

            log.info("[{}]缓冲区满，缓冲区大小：{}，缓冲区开始接收数据时间:{}, 缓冲区满花费时间:{}ms", bufferIndex,
                    fulledBuffer.size(),
                    Util.uTCMilliSecondsToDateStringWithMs(bufferFullTimeCalculator.getStartTime()),
                    bufferFullTimeCalculator.getSpendTime());

            // 添加任务
            diskTaskBuffer.put(new DiskTask(fulledBuffer, bufferIndex, walId, System.currentTimeMillis()));

            bufferIndex++;
            bufferFullTimeCalculator.begin();
        }
    }

    /**
     * 保存一组报文
     *
     * @param fulledBuffer
     * @param bufferId
     * @param walId
     * @return
     */
    public boolean saveFulledPackageBuffer(List<DataResponsePackage> fulledBuffer, long bufferId, long walId) {
        log.info("[{}]保存一组报文，报文组大小：{}，开始保存", bufferId, fulledBuffer.size());

        SavePackageResult savePackageResult = saveMultiDataResponsePackage(fulledBuffer, bufferId, walId);

        log.info("[{}]保存一组报文数据至磁盘:{}，至tsdb队列:{}", bufferId, savePackageResult.saveDiskSuccess,
                savePackageResult.saveTsdbSuccess);

        if (statisticService.getShowStatisticAfterInsertData()) {
            log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
        }

        return savePackageResult.isSaveDiskSuccess();

    }

    /**
     * 保存多个报文到磁盘，并建立TSDB任务，报文数等于MaxItem数
     *
     * @param packages 待保存报文，报文数等于MaxItem数
     * @return
     */
    public SavePackageResult saveMultiDataResponsePackage(List<DataResponsePackage> packages, long bufferId, Long walId) {

        SavePackageResult savePackageResult = new SavePackageResult();

        if (packages.size() != Globals.MAX_LOG_ITEM) {
            log.error("处理报文数{}不等于MaxItem大小", packages.size());
            return savePackageResult;
        }

        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create().begin();

        log.info("[{}]开始整理元数据", bufferId);
        // 转换一组数据包为一组硬盘的LogItem，计算总rawPacket大小，得到一组tsdb的tags
        List<LogItem> logItems = new ArrayList<>(packages.size());
        List<DBVal> dbVals = new ArrayList<>(packages.size());

        int totalDataSize = 0;
        for (DataResponsePackage dataResponsePackage : packages) {
            logItems.add(new LogItem(LogItem.DEFAULT_LOG_ITEM_TYPE, dataResponsePackage.getTimestamp(),
                    dataResponsePackage.getRawPacket()));
            totalDataSize += dataResponsePackage.getRawPacket().length;

            HashMap<String, String> tags = new HashMap<>();
            tags.put("sMac", dataResponsePackage.getSMac());
            tags.put("dMac", dataResponsePackage.getDMac());
            tags.put("sPort", dataResponsePackage.getSPort());
            tags.put("dPort", dataResponsePackage.getDPort());
            tags.put("sIp", dataResponsePackage.getSIp());
            tags.put("dIp", dataResponsePackage.getDIp());
            tags.put("probeId", String.valueOf(dataResponsePackage.getProbeId()));
            Point point = new Point(metricName, tags);

            DBVal val = new DBVal();
            val.setPoint(point);
            val.setUtcTimeMilliSeconds(dataResponsePackage.getTimestamp());
            dbVals.add(val);
        }

        log.info("[{}]开始保存至磁盘", bufferId);
        List<Double> doubleIndexes;

        try {
            // 已做转换，将long转换为double
            doubleIndexes = logFileService.appendLogs(logItems, totalDataSize, walId);
        } catch (Exception e) {
            log.error("[{}]写入数据到磁盘失败", bufferId, e);
            return savePackageResult;
        }
        spentTimeCalculator.end();
        int insertCount = doubleIndexes.size();
        long logFileId = Double.doubleToRawLongBits(doubleIndexes.get(0)) / Globals.MAX_LOG_ITEM;
        log.info("[{}]写入磁盘成功，写入磁盘花费总时间:{}ms 插入{}条", bufferId, spentTimeCalculator.getSpendTime(),
                insertCount);
        statisticService.increDataCountDiskSuccess(insertCount);
        savePackageResult.setSaveDiskSuccess(true);

        for (int i = 0; i < insertCount; i++) {
            dbVals.get(i).setValue(doubleIndexes.get(i));
        }

        tsdbTaskBuffer.put(new TsdbTask(dbVals, bufferId, logFileId, System.currentTimeMillis()));
        log.info("生成TSDB任务，放入阻塞队列，任务共有{}条，logFile Id = {}", dbVals.size(), logFileId);
        savePackageResult.setSaveTsdbSuccess(true);
        return savePackageResult;
    }

    @Deprecated
    public void saveSingleDataResponsePackage(DataResponsePackage dataResponsePackage) {
        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create().begin();
        log.debug("开始处理探针上传的数据包:{}", dataResponsePackage);
        statisticService.incrDataCount();


        log.debug("开始保存至磁盘");
        // 封装 LogItem
        LogItem logItem = new LogItem((short) 0x55, dataResponsePackage.getTimestamp(), dataResponsePackage.getRawPacket());

        long longIndex;
        try {
            longIndex = logFileService.appendLog(logItem);
            spentTimeCalculator.end();
        } catch (Exception e) {
            e.printStackTrace();
            log.error("写入数据到磁盘失败", e);
            return;
        }
        statisticService.increDataCountDiskSuccess();
        double index = Double.longBitsToDouble(longIndex);
        log.debug("写入磁盘成功，接收探针上传数据到写入磁盘花费总时间:{}ms 插入{}条", spentTimeCalculator.getSpendTime(), 1);

        // 写入数据库
        log.debug("开始尝试插入ResoDB，longIndex={}，doubleIndex={}", longIndex, index);

        HashMap<String, String> tags = new HashMap<>();
        tags.put("sMac", dataResponsePackage.getSMac());
        tags.put("dMac", dataResponsePackage.getDMac());
        tags.put("sPort", dataResponsePackage.getSPort());
        tags.put("dPort", dataResponsePackage.getDPort());
        tags.put("sIp", dataResponsePackage.getSIp());
        tags.put("dIp", dataResponsePackage.getDIp());
        tags.put("probeId", String.valueOf(dataResponsePackage.getProbeId()));
        Point point = new Point(metricName, tags);

        long[] utcTimes = new long[1];
        double[] values = new double[1];

        utcTimes[0] = dataResponsePackage.getTimestamp();
        values[0] = index;
        PointVals pointVals = new PointVals(point, 1, utcTimes, values);
        try {
            int insertCount = entry.insertSinglPoint(pointVals);

            spentTimeCalculator.end();

            if (insertCount != 0) {
                statisticService.increDataCountTsdbSuccess();
                log.debug("插入ResoDB结束，point={}，插入{}条", pointVals, insertCount);
                log.debug("接收探针上传数据到插入时序数据库花费总时间:{}ms 插入{}条", spentTimeCalculator.getSpendTime(), insertCount);
            } else {
                log.error("接收探针上传数据到插入时序数据库花费总时间:{}ms 插入失败", spentTimeCalculator.getSpendTime());
            }

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
            }
        } catch (Exception e) {
            log.error(String.format("插入ResoDB失败，ip:%s port:%d", entry.getHost(), entry.getPort()), e);

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
            }

        }
    }

    private void checkDiskQueue() {
        while (true) {
            try {
                DiskTask diskTask = diskTaskBuffer.take();
                log.info("[{}]从磁盘任务阻塞队列中取出任务，开始保存至磁盘", diskTask.getBufferId());
                if (!saveFulledPackageBuffer(diskTask.getPackages(), diskTask.getBufferId(), diskTask.getWalId())) {
                    // 保存硬盘失败，重新放回
                    log.warn("[{}]保存到磁盘失败，重新放回磁盘任务队列", diskTask.bufferId);
                    diskTask.setExecTime(System.currentTimeMillis() + retryTaskDelayMilli);
                    diskTaskBuffer.put(diskTask);
                }
            } catch (InterruptedException e) {
                log.error("从磁盘任务阻塞队列中取出任务失败", e);
            }
        }
    }

    private void checkTsdbQueue() {
        while (true) {
            try {
                TsdbTask tsdbTask = tsdbTaskBuffer.take();
                log.info("[{}]从阻塞队列中取出任务，开始保存索引至TSDB，logFile Id = {}", tsdbTask.bufferId, tsdbTask.getLogId());
                logFileService.markTsdbTryStamp(tsdbTask.getLogId());
                if (saveToTsdb(tsdbTask.getDbVals(), tsdbTask.bufferId)) {
                    // 保存tsdb存储状态
                    logFileService.markTsdbSuccess(tsdbTask.getLogId());
                    // 清除wal文件
                    if(enableWal) {
                        Long walStamp = logFileService.getWalStamp(tsdbTask.getLogId());
                        walFileService.clear(walStamp);
                    }
                } else {
                    // 保存tsdb失败，重新放回
                    log.warn("[{}]保存到tsdb失败，重新放回tsdb任务队列", tsdbTask.bufferId);
                    tsdbTask.setExecTime(System.currentTimeMillis() + retryTaskDelayMilli);
                    tsdbTaskBuffer.put(tsdbTask);
                }
            } catch (InterruptedException e) {
                log.error("从tsdb队列中取出任务失败", e);
            }
        }
    }

    /**
     * 将数据存入TSDB
     *
     * @return
     */
    private boolean saveToTsdb(List<DBVal> dbVals, Long bufferId) {
        log.info("[{}]开始尝试插入TSDB", bufferId);
        SpentTimeCalculator spentTimeCalculator = SpentTimeCalculator.create().begin();

        int insertCount = dbVals.size();

        boolean saveResult = false;

        try {
            int insertRes = entry.insertMultiPoint(dbVals);
            log.debug("insertRes={}", insertRes);
            spentTimeCalculator.end();

            if (insertRes != -1) {
                statisticService.increDataCountTsdbSuccess(insertCount);
                saveResult = true;
//                log.debug("插入ResoDB结束，point={}，插入{}条", dbVals, insertCount);
                log.info("插入时序数据库花费时间:{}ms 插入{}条", spentTimeCalculator.getSpendTime(), insertCount);
            } else {
                log.error("插入时序数据库花费时间:{}ms 插入失败", spentTimeCalculator.getSpendTime());
            }

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("到目前为止的统计信息：{}", statisticService.getStatistic().toString());
            }
        } catch (Exception e) {
            log.error("[{}]插入ResoDB失败，ip:{} port:{}", bufferId, entry.getHost(), entry.getPort(), e);

            if (statisticService.getShowStatisticAfterInsertData()) {
                log.info("[{}]到目前为止的统计信息：{}", bufferId, statisticService.getStatistic().toString());
            }
            return false;
        }
        return saveResult;
    }


    public void handleScanResponsePackage(ScanResponsePackage scanResponsePackage) {
        log.debug("开始处理数据包ScanResponsePackage:{}", scanResponsePackage);

        // 保存至探针信息
        short probeId = scanResponsePackage.getId();
        if (!probeService.getProbes().containsKey(probeId)) {
            probeService.getProbes().put(probeId, new Probe());
        }
        probeService.getProbes().get(probeId).updateFromScanResponsePackage(scanResponsePackage);

        log.debug("成功处理数据包{}", scanResponsePackage);
    }

    public void handleListenResponsePackage(ListenResponsePackage packet) {
        log.debug("开始处理数据包ListenResponsePackage: {}", packet);
        log.debug("成功处理数据包{}", packet);
    }


}
