package db.netty;

import db.Globals;
import db.entity.packages.*;
import db.util.ConvertUtils;
import io.netty.buffer.ByteBuf;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.socket.DatagramPacket;
import io.netty.handler.codec.MessageToMessageDecoder;
import lombok.extern.slf4j.Slf4j;

import java.util.List;

@Slf4j
public class Decoder extends MessageToMessageDecoder<DatagramPacket> {
    private boolean showRecvBytes = true;

    @Override
    protected void decode(ChannelHandlerContext ctx, DatagramPacket datagramPacket, List<Object> out) throws Exception {
        ByteBuf udpPacketBytes = datagramPacket.content();
        byte[] bytesReady = new byte[udpPacketBytes.readableBytes()];
        udpPacketBytes.getBytes(0, bytesReady);

        if(this.showRecvBytes) {
            log.debug("解码器收到Host:{} port:{}发送的数据：{} 从 {}",
                    datagramPacket.sender().getHostString(),
                    datagramPacket.sender().getPort(),
                    ConvertUtils.bytes2HexArrayWithIndex(bytesReady), ctx);
            log.debug("解码器收到Host:{} port:{}发送的数据：{} 从 {}",
                    datagramPacket.sender().getHostString(),
                    datagramPacket.sender().getPort(),
                    ConvertUtils.bytes2HexArray(bytesReady), ctx);
        }

        int magic = udpPacketBytes.getInt(0);

        if (magic != Globals.PROBE_MAGIC) {
            log.warn("魔数错误，消息为：{}", ConvertUtils.bytes2HexArray(bytesReady));
            return;
        }

        int ver = udpPacketBytes.getInt(4);
        short operationType = udpPacketBytes.getShort(8);
        short responseCode = udpPacketBytes.getShort(10);
        int packageId = udpPacketBytes.getInt(12);
        short operationCode = udpPacketBytes.getShort(12);
        short packageCode = udpPacketBytes.getShort(14);
        short packageLength = udpPacketBytes.getShort(16);

        BasePackage basePackage;

        if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_SCAN) {
            log.debug("解码器转换为设备扫描请求报文");
            basePackage = new ScanRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_SCAN) {
            log.debug("解码器转换为设备扫描响应报文");
            basePackage = new ScanResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_LINK) {
            log.debug("解码器转换为设备链接请求报文");
            basePackage = new ScanRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_LINK) {
            log.debug("解码器转换为设备链接响应报文");
            basePackage = new ScanResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_BASE_INFO) {
            log.debug("解码器转换为基本信息配置请求报文");
            basePackage = new BaseInfoWriteRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_BASE_INFO) {
            log.debug("解码器转换为基本信息配置响应报文");
            basePackage = new BaseInfoWriteResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_TABLE) {
            log.debug("解码器转换为安全表项配置请求报文");
            basePackage = new TableInfoWriteRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_TABLE) {
            log.debug("解码器转换为安全表项配置响应报文");
            basePackage = new TableInfoWriteResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_TABLE) {
            log.debug("解码器转换为安全表项读取请求报文");
            basePackage = new TableInfoReadRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_TABLE) {
            log.debug("解码器转换为安全表项读取响应报文");
            basePackage = new TableInfoReadResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_STRATEGY) {
            log.debug("解码器转换为安全策略配置请求报文");
            basePackage = new StrategyWriteRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_STRATEGY) {
            log.debug("解码器转换为安全策略配置响应报文");
            basePackage = new StrategyWriteResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_STRATEGY) {
            log.debug("解码器转换为安全策略读取请求报文");
            basePackage = new StrategyReadRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_STRATEGY) {
            log.debug("解码器转换为安全策略读取响应报文");
            basePackage = new StrategyReadResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_LIMIT) {
            log.debug("解码器转换为广播限流配置请求报文");
            basePackage = new LimitWriteRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_LIMIT) {
            log.debug("解码器转换为广播限流配置响应报文");
            basePackage = new LimitWriteResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_LIMIT) {
            log.debug("解码器转换为广播限流读取请求报文");
            basePackage = new LimitReadRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_LIMIT) {
            log.debug("解码器转换为广播限流读取响应报文");
            basePackage = new LimitReadResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_CLIENT_INFO) {
            log.debug("解码器转换为接入终端信息读取请求报文");
            basePackage = new ClientReadRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_CLIENT_INFO) {
            log.debug("解码器转换为接入终端信息读取响应报文");
            basePackage = new ClientReadResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && (operationCode == Globals.OPERATION_LISTEN_LOGIN ||
                operationCode == Globals.OPERATION_LISTEN_CANCEL)) {
            log.debug("解码器转换为监听登录/取消请求报文");
            basePackage = new ListenRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && (operationCode == Globals.OPERATION_LISTEN_LOGIN ||
                operationCode == Globals.OPERATION_LISTEN_CANCEL)) {
            log.debug("解码器转换为监听登录/取消响应报文");
            basePackage = new ListenResponsePackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == 0x0006) {
            log.debug("解码器转换为监听数据报文");
            basePackage = new DataResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_LOCK) {
            log.debug("解码器转换为系统锁定配置请求报文");
            basePackage = new LockWriteRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_WRITE &&
                packageCode == Globals.PACKAGE_LOCK) {
            log.debug("解码器转换为系统锁定配置响应报文");
            basePackage = new LockWriteResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_LOCK) {
            log.debug("解码器转换为系统锁定读取请求报文");
            basePackage = new LockReadRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_READ &&
                packageCode == Globals.PACKAGE_LOCK) {
            log.debug("解码器转换为系统锁定读取响应报文");
            basePackage = new LockReadResponsePackage();
        } else if (operationType == Globals.TYPE_REQUEST && operationCode == Globals.OPERATION_SAVE_CONFIG) {
            log.debug("解码器转换为保存配置信息请求报文");
            basePackage = new SaveConfigRequestPackage();
        } else if (operationType == Globals.TYPE_RESPONSE && operationCode == Globals.OPERATION_SAVE_CONFIG) {
            log.debug("解码器转换为保存配置信息响应报文");
            basePackage = new SaveConfigResponsePackage();
        } else {
            log.error("解码器无法识别报文，operationType={}，operationCode={}，packageCode={}", operationType, operationCode,
                    packageCode);
            return;
        }

        // 把byteBuf中的内容填充到包对象中
        basePackage.setHeader(udpPacketBytes);
        basePackage.setPayload(udpPacketBytes);

        basePackage.setFromHost(datagramPacket.sender().getHostString());
        basePackage.setFromPort(datagramPacket.sender().getPort());

        out.add(basePackage);
        if(this.showRecvBytes) {
            log.info("解码器转码Host:{} port:{}发送的数据为:{}",
                    datagramPacket.sender().getHostString(),
                    datagramPacket.sender().getPort(),
                    basePackage);
        }
    }

    public void setShowRecvBytes(boolean showRecvBytes) {
        this.showRecvBytes = showRecvBytes;
    }
}


