package db.service;

import db.*;
import db.entity.LogFileMeta;
import db.mapper.LogFileMetaMapper;
import db.util.Util;
import lombok.extern.slf4j.Slf4j;
import org.pcap4j.core.*;
import org.pcap4j.packet.UnknownPacket;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.io.IOException;
import java.text.ParseException;
import java.util.*;

@Service
@Slf4j
public class QueryService {

    @Autowired
    private LogFileService logFileService;

    @Autowired
    private LogFileMetaMapper logFileMetaMapper;

    @Value("${reso-db.host}")
    private String host;

    @Value("${reso-db.port}")
    private int port;

    @Value("${reso-db.metric-name}")
    private String metricName;

    @Value("${server.port}")
    private String serverPort;

    private DBApiEntry entry;

    @PostConstruct
    public void afterCons() {
        this.entry = DBApiEntry.initApiEntry(host, port);
    }

    public String queryByTimeAndCondition(String start, String end, Map<String, String> tags) throws ParseException,
            IOException, PcapNativeException, NotOpenException {

        Point point = new Point(metricName, tags);

        long startTime = Util.dateStringToUTCMilliSeconds(start);
        long endTime = Util.dateStringToUTCMilliSeconds(end);

        DBApiEntry entry = DBApiEntry.initApiEntry(host, port);

        List<DBVal> lists = entry.getHistRaw(point, startTime, endTime);
        log.info("queryByTimeAndCondition，查询条件为：start:{}, end:{}, point:{}", startTime, endTime, point);
        log.info("查询ResoDB，查询到：{}", lists);
        // 根据list查对应报文位置
        List<Long> locs = new ArrayList<>();
        for (DBVal dbVal : lists) {
            locs.add(Double.doubleToRawLongBits(dbVal.getValue()));
        }
        Collections.sort(locs);

        PcapHandle handle = Pcaps.openOffline("./pcap/template.pcap");
        String pcapName = "" + startTime + endTime + System.currentTimeMillis();
        PcapDumper dumper = handle.dumpOpen("./pcap/" + pcapName + ".pcap");

        for (Long longIndex : locs) {
            byte[] logItem = logFileService.queryLog(longIndex);
            log.info("查询磁盘，查询到：{}", logItem);
            UnknownPacket packet = UnknownPacket.newPacket(logItem, 0, logItem.length);
            dumper.dump(packet);
        }

        dumper.close();
        handle.close();

        StringBuilder returnLink = new StringBuilder();
        returnLink.append("http://").append("127.0.0.1").append(":")
                .append(serverPort).append("/download?pcapName=").append(pcapName);

        log.info("返回链接：{}", returnLink);
        return String.valueOf(returnLink);
    }

    public String queryByTime(String start, String end) throws ParseException, IOException, PcapNativeException, NotOpenException {

        long startTime = Util.dateStringToUTCMilliSeconds(start);
        long endTime = Util.dateStringToUTCMilliSeconds(end);

        log.info("queryByTime，查询条件为：start:{}, end:{}", startTime, endTime);

        List<LogFileMeta> logFileMetas = logFileMetaMapper.getValidLogFileMetaBetweenTime(startTime, endTime);
        log.info("queryByTime，查询B+树，查询到：{}", logFileMetas);

        PcapHandle handle = Pcaps.openOffline("./pcap/udp.pcap");
        String pcapName = "" + startTime + endTime + System.currentTimeMillis();
        PcapDumper dumper = handle.dumpOpen("./pcap/" + pcapName + ".pcap");

        for (LogFileMeta logFileMeta : logFileMetas) {
            List<byte[]> bytes = logFileService.queryLogBetweenTime(logFileMeta, startTime, endTime);

            for (byte[] item : bytes) {
                UnknownPacket packet = UnknownPacket.newPacket(item, 0, item.length);
                dumper.dump(packet);
            }
        }
        dumper.close();
        handle.close();


        StringBuilder returnLink = new StringBuilder();
        returnLink.append("http://").append("127.0.0.1").append(":")
                .append(serverPort).append("/download?pcapName=").append(pcapName);

        log.info("返回链接：{}", returnLink);
        return String.valueOf(returnLink);
    }
}
